"""
`Importable` is a prototype that can be attached to any ViUR module to provide a configurable
and and hookable interface for executing imports, partly automatically.
"""

import base64
import logging
import time
import types
import typing as t
from datetime import datetime as dt

from google.cloud.datastore import _app_engine_key_pb2

from viur.core import conf, current, db, email, errors, utils
from viur.core.decorators import exposed
from viur.core.prototypes.tree import SkelType
from viur.core.skeleton import SkeletonInstance
from viur.core.tasks import CallDeferred
from .importer import Importer

logger = logging.getLogger(__name__)

JINJA_EMAIL_TEMPLATE = """{{ skel["targetportal"] }}.{{ skel["targetmodule"] }} import done, total:{{ skel["total"]}}, updated:{{ skel["updated"]}}, removed:{{ skel["removed"]}}
<!DOCTYPE html>
<html lang="de">
<body>
    Der Import von {{ skel["sourceportal"] }}.{{ skel["sourcemodule"] }}
    nach {{ skel["targetportal"] }}.{{ skel["targetmodule"] }} wurde erfolgreich durchgeführt.<br>

    Es wurden {{ skel["total"]}} Datensätze geprüft, davon {{ skel["updated"] }} aktualisiert.<br>
    {{ skel["removed"] }} Datensätze wurden gelöscht.
</body>
</html>"""


class _AppKey(db.Key):
    def to_legacy_urlsafe(self, project_id: str | None = None) -> bytes:
        """
        Converts this key into the (urlsafe) protobuf string representation.
        :return: The urlsafe string representation of this key
        """
        currentKey = self
        pathElements: list[_app_engine_key_pb2.Path.Element] = []
        while currentKey:
            pathElements.insert(
                0,
                _app_engine_key_pb2.Path.Element(
                    type=currentKey.kind,
                    id=currentKey.id,
                    name=currentKey.name,
                ),
            )
            currentKey = self.parent

        if not project_id:
            project_id = conf.project_id

        reference = _app_engine_key_pb2.Reference(
            app=project_id,
            path=_app_engine_key_pb2.Path(element=pathElements),
        )
        raw_bytes = reference.SerializeToString()
        return base64.urlsafe_b64encode(raw_bytes).strip(b"=")


class Importable:
    """
    Importable prototype

    This prototype can be attached to any module making it importable. Data imports can both be done between ViUR2
    and ViUR3 systems. Only the JSON interface is used.

    The module then requires for a *import_conf* to be defined (see below) and provides several functions for triggering
    and maintenance. The prototype is used in combination with the importer module from the root.
    """

    # !!!TODO!!! Should be turned into a class!
    import_conf: dict[str, t.Any] = {
        "source": {
            "url": None,  # this is mandatory!
            "auth": "viur",  # It can also be other sources used
            "method": "userpassword"  # TODO: Document all types
        },
        "module": None,  # Source module; if None, same name as the importable module will be used.
        "translate": None,  # Translation dictionary; if omitted, it will be generated from bone:bone
        "translate.update": None,  # Translation dictionary; extend and automatically generated one (if translate==None)
        "translate.ignore": None,  # Bones to be ignored in automatically generated translation
        "action": "list",  # Action to run, default is "list"
        "limit": 99,  # Amount of items to fetch per request (only for list-ables)
        "render": "vi",  # Renderer to run on
        "params": None,  # Further parameters passed to the action
        "follow": [],  # Following modules to be imported, that depend on this import.
        "filter": {},  # Define a custom filter to be used for preparation and cleaning
        "updateRelations": True,  # Allows to disable update relation tasks enforcements when configured to False.
        "enforce": False,  # Enforce all full skeletons to be rewritten.
        "skip": lambda values: False,  # don't skip any entry by default
        "inform": False,  # either an e-mail address to inform, or True for current user, False otherwise.
        "clear": True,  # Do do_clear and delete not imported entries
    }

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        super().__init__(*args, **kwargs)
        self._bone_translation_table: dict[str, dict] = {}
        """the final translation table once created by create_config()"""

    def modify_skel_key(
        self,
        key: db.Key,
        skel: SkeletonInstance,
        values: dict[str, t.Any],
        translate: dict[str, str | t.Callable],
        source_key: str,
    ) -> db.Key:
        return key

    @staticmethod
    def translate_key(skel: SkeletonInstance, bone: str, value: t.Any, **_: t.Any) -> int:
        """
        Helper function to replace a key during translation, even if it's not a keyBone.
        """
        if not value:
            return 0
        try:
            key = db.Key.from_legacy_urlsafe(value)
            assert not key.parent, "Not implemented"
        except Exception:
            logger.exception("Can't convert key")
            return 0

        if skel[bone] != key:
            skel[bone] = key
            return 1

        return 0

    @staticmethod
    def translate_select_values(skel: SkeletonInstance, bone_name: str, value: t.Any, matching: dict) -> int:
        """
        Helper to rewrite select values from a given matching table.
        """
        is_list = True
        if not isinstance(value, list):
            value = [value]
            is_list = False

        changes = 0
        new_value = []

        for val in value:
            if val in matching:
                val = matching[val]
            new_value.append(val)

        if not skel[bone_name] or set(skel[bone_name]) != set(new_value):
            if not is_list:
                skel.setBoneValue(bone_name, new_value[0])
            else:
                skel.setBoneValue(bone_name, new_value)

            changes += 1

        return changes

    def get_handler(self) -> str:
        admin_info = self.describe()
        assert (handler := admin_info.get("handler"))
        return handler.split(".", 1)[0]

    def import_skel(self, skelType: SkelType = None) -> SkeletonInstance:
        handler = self.get_handler()

        if handler == "tree":
            try:
                return self.editSkel(skelType=skelType)
            except:
                pass

        return self.editSkel()

    @exposed
    def start_import(
        self,
        *,
        follow: bool = False,
        enforce: bool = False,
        inform: str = "false",
        dry_run: bool = False,
        otp: str | None = None,
        debug: bool = False,
        import_conf_name: str = "import_conf",
        **kwargs: t.Any,
    ) -> str:
        cuser = current.user.get()
        if not cuser or "root" not in cuser["access"]:
            raise errors.Unauthorized()

        # Additional credentials
        import_conf = getattr(self, import_conf_name)
        import_conf = import_conf() if callable(import_conf) else import_conf

        source = import_conf.get("source") or {}

        if source.get("type") == "viur" and source.get("auth") == "userpassword+otp":
            # Hint for OTP
            if not otp:
                raise errors.BadRequest("Requires 'otp' key to succeed")

            source["otp"] = otp

        if self.get_handler() == "tree":
            kwargs.setdefault("skelType", "node")
            kwargs.setdefault("_autoSkelType", True)

        if "@" not in inform:
            inform = utils.parse.bool(inform)
            inform = import_conf.get("inform") or inform

        inform = inform if isinstance(inform, str) else cuser["name"] if inform else None

        self.do_import(
            inform=inform,
            follow=follow,
            enforce=enforce,
            dry_run=dry_run,
            source=source or None,
            debug=debug,
            import_conf_name=import_conf_name,
            _queue="import",
            **kwargs,
        )

        info = f"Import of {self.moduleName} kicked off"
        if inform:
            info += f", {inform} will be notified by e-mail on finish"

        return info

    @CallDeferred
    def do_import(
        self,
        importdate: dt | None = None,
        inform: str | None = None,
        spawn: int = 0,
        cursor: str | None = None,
        total: int = 0,
        updated: int = 0,
        follow: bool = False,
        enforce: bool = False,
        dry_run: bool = False,
        cookies: dict | None = None,
        source: dict | None = None,
        delete_filter: dict | None = None,
        debug: bool = False,
        import_conf_name: str = "import_conf",
        **kwargs: t.Any,
    ) -> None | bool:
        import_conf = getattr(self, import_conf_name)
        import_conf = import_conf() if callable(import_conf) else import_conf
        assert import_conf.get("source"), "No source specified to import from"

        # Mark this request as an import task
        current.request.get().kwargs["isImportTask"] = True  # FIXME!
        if importdate is None:
            importdate = utils.utcNow().replace(microsecond=0)  # need to remove microseconds!!!

        logger.debug(f"{self.moduleName!r} {importdate=} {total=}")

        # Login
        imp = Importer(
            (import_conf.get("source") or {}) | (source or {}),
            render=import_conf.get("render", "vi"),
            cookies=cookies,
        )

        # In case of a hierarchy, always assume skelType "node"
        handler = self.get_handler()
        if handler == "tree":
            kwargs.setdefault("skelType", "node")

        if not kwargs:
            params = {}
        else:
            params = kwargs.copy()

        if conf_params := import_conf.get("params"):
            if callable(conf_params):
                conf_params = conf_params(cursor)

            assert isinstance(
                params, dict
            ), "import_conf[\"params\"] must be either 'dict' or 'callable' returning dict"
            params.update(conf_params)

        if "limit" not in params:
            params["limit"] = import_conf.get("limit", 99)
            params["amount"] = import_conf.get("limit", 99)

        if "cursor" not in params:
            params["cursor"] = cursor

        url = f"""{import_conf.get("module", self.moduleName)}/{import_conf.get("action", "list")}"""
        answ = imp.post(url, params=params, timeout=60)

        if not answ.ok:
            logger.error(f"Cannot fetch list from {url=}, {answ.status_code=}")
            raise errors.BadRequest()

        try:
            answ = answ.json()

        except:
            logger.warning(
                "Target module %r does not exist in source or some other error occured - skipping",
                import_conf.get("module", self.moduleName),
            )
            self._kickoff_follow(importdate, inform, import_conf_name=import_conf_name, **kwargs)
            return None

        # Get skeleton
        skel = self.import_skel(skelType=kwargs.get("skelType"))

        self.create_config(skel)

        # Perform import
        if isinstance(answ, dict):
            skellist = answ.get("skellist", [])
            cursor = answ.get("cursor")
        elif isinstance(answ, list):
            skellist = answ
        else:
            raise NotImplementedError(f"Cannot handle type {type(answ)}")

        for values in skellist:
            total += 1

            if "skip" in import_conf and import_conf["skip"](values):
                continue

            # logger.debug(f"{values=}")

            if self._convert_entry(
                imp,
                skel,
                values,
                importdate,
                skel_type=kwargs.get("skelType"),
                enforce=enforce,
                dry_run=dry_run,
                updateRelations=import_conf.get("update_relations", True),
                debug=debug,
                import_conf_name=import_conf_name,
            ):
                updated += 1

                # if total >= 5:
                #    skellist = ()
                #    break

        logger.info("%s: %d entries imported, %d entries updated", self.moduleName, total, updated)

        if not skellist or cursor is None:
            imp.logout()  # log-out now, as we're finished reading

            # Clear deleted entries?
            if import_conf.get("clear", True) and "importdate" in skel:
                self.do_clear(
                    importdate,
                    inform,
                    total,
                    updated,
                    follow=follow,
                    dry_run=dry_run,
                    delete_filter=delete_filter,
                    import_conf_name=import_conf_name,
                    _queue="import",
                    **kwargs,
                )
                return None

            logger.info(
                "%s: Import finished; %d entries in total, %d updated",
                self.moduleName,
                total,
                updated,
            )

            if inform:
                email.sendEMail(
                    dests=inform,
                    stringTemplate=JINJA_EMAIL_TEMPLATE,
                    skel={
                        "sourceportal": import_conf["source"]["url"],
                        "targetportal": conf.instance.project_id,
                        "sourcemodule": import_conf.get("module", self.moduleName),
                        "targetmodule": self.moduleName,
                        "total": total,
                        "updated": updated,
                        "removed": 0,
                    },
                )

            self.onImportFinished(self.moduleName, total, updated)

            if follow:
                self._kickoff_follow(importdate, inform, **kwargs)

            return True

        self.do_import(
            importdate=importdate,
            inform=inform,
            spawn=spawn + 1,
            cursor=cursor,
            total=total,
            updated=updated,
            follow=follow,
            dry_run=dry_run,
            cookies=imp.cookies.get_dict(),
            delete_filter=delete_filter,
            import_conf_name=import_conf_name,
            _queue="import",
            **kwargs,
        )

        return True

    def import_generate_translation(self, skel: SkeletonInstance) -> dict[str, t.Any]:
        """
        Automatically generates a 1-to-1 translation from the given skel.
        Can be subclasses for custom behavior.
        """
        tr = {k: k for k in skel.keys()}

        if "parentrepo" in skel:
            tr["parentrepo"] = Importable.translate_key

        if "parententry" in skel:
            tr["parententry"] = Importable.translate_key

            # ViUR2 legacy bullshit, in trees parententry was called parentdir...
            tr["parentdir"] = lambda bone, **kwargs: Importable.translate_key(
                bone="parententry", **kwargs
            )

        return tr

    def create_config(self, skel: SkeletonInstance, import_conf_name: str = "import_conf") -> None:
        import_conf = getattr(self, import_conf_name)
        import_conf = import_conf() if callable(import_conf) else import_conf

        # Get translation from config
        tr = import_conf.get("translate")

        # If it's a 1-to-1 list, make it a dict.
        if isinstance(tr, list):
            tr = {k: k for k in tr}

        # Otherwise, when not specified, automatically construct a translation from the skeleton with some specials
        elif tr is None:
            tr = self.import_generate_translation(skel)

        # update further bones to (probably automatic) translation
        tr |= import_conf.get("translate.update") or {}

        # Remove any bones that should be ignored
        for k in list(tr.keys()):
            if k in ["key", "changedate", "importdate"] + (
                import_conf.get("translate.ignore") or []
            ):
                del tr[k]

        assert isinstance(tr, dict), "translation must be specified as dict!"
        self._bone_translation_table[import_conf_name] = tr

    def do_import_entry(
        self,
        key: str,
        module: str | None = None,
        kindName: str | None = None,
        skel_type: SkelType = "node",
        debug: bool = False,
        import_conf_name: str = "import_conf",
    ) -> bool | None:
        import_conf = getattr(self, import_conf_name)
        import_conf = import_conf() if callable(import_conf) else import_conf

        importdate = utils.utcNow()
        source = import_conf.get("source") or {}

        # Login
        try:
            imp = Importer(
                source=source,
                render=import_conf.get("render", "vi")
            )

        except Exception as e:
            logger.exception(e)
            return None

        skel = self.import_skel(skelType=skel_type)

        self.create_config(skel, import_conf_name=import_conf_name)

        key = db.KeyClass.from_legacy_urlsafe(key)

        if not kindName:
            kindName = import_conf.get("module")

        key = _AppKey(kindName, key.id_or_name)  # type: ignore[attr-defined]

        project_id = source.get("project_id")
        assert project_id, f"Please set the project_id of portal {source.get('url')}"
        key = key.to_legacy_urlsafe(project_id=project_id).decode("utf-8")

        if not module:
            module = import_conf.get("module", self.moduleName)
        url = f"{module}/view/{key}"

        answ = imp.post(url, timeout=60)
        if not answ.ok:
            logger.error(
                "Cannot fetch list from %r, error %d occured", url, answ.status_code
            )
            raise errors.BadRequest()

        try:
            answ = answ.json()
        except Exception:
            logger.warning(
                "Target module %r does not exist in source or some other error occured - skipping",
                import_conf.get("module", self.moduleName),
            )
            return None
        assert isinstance(answ, dict)
        values = answ["values"]

        return self._convert_entry(
            imp,
            skel,
            values,
            importdate,
            skel_type,
            enforce=import_conf.get("enforce", False),
            updateRelations=import_conf.get("updateRelations", True),
            debug=debug,
            import_conf_name=import_conf_name,
        )

    def _convert_entry(
        self,
        imp: "Importer",
        skel: SkeletonInstance,
        values: dict[str, t.Any],
        importdate: dt,
        skel_type: SkelType = None,
        enforce: bool = False,
        dry_run: bool = False,
        updateRelations: bool = True,
        debug: bool = False,
        import_conf_name: str = "import_conf",
    ) -> bool:
        """
        Internal function for converting one entry.
        """
        skel.setEntity(db.Entity())
        ret = imp.values_to_skel(
            skel,
            values,
            self._bone_translation_table[import_conf_name],
            source_key="key" if "key" in values else "id",  # ViUR 1.x
            enforce=enforce,
            debug=debug,
        )

        if "outdated" in skel:
            skel["outdated"] = False

        if dry_run:
            logger.info(f"dry run {ret=}, {skel=}")
            return ret != 0

        if ret != 0 or enforce:
            if not self.onEntryChanged(skel, values):
                return False

            # Set importdate when available
            if "importdate" in skel:
                logger.debug(
                    "%s: Setting importdate on %r to %r",
                    self.moduleName,
                    skel["key"],
                    importdate,
                )
                skel["importdate"] = importdate

            for attempt in (rng := range(3)):  # TODO: Make it configureable
                try:
                    assert skel.toDB(update_relations=updateRelations)
                except Exception as e:
                    logger.exception(f"cannot convert {skel['key']} : {e!s}   {skel!r}")

                    if attempt == rng.stop - rng.step:
                        raise e

                    logger.info(f"Waiting {2 ** attempt} seconds for the next attempt")
                    time.sleep(2 ** attempt)

                else:
                    break

            handler = self.get_handler()

            if handler in ["hierarchy", "tree"]:
                if ret < 0:
                    self.onAdded(skel_type, skel)
                else:
                    self.onEdited(skel_type, skel)
            else:
                if ret < 0:
                    self.onAdded(skel)
                else:
                    self.onEdited(skel)

            return True
        else:
            if not self.onEntryUnchanged(skel, values):
                return False

            # Save with importdate when required
            if "importdate" in skel:
                skel["importdate"] = importdate

                assert skel.toDB(update_relations=True)
        return False

    def onEntryChanged(self, skel: SkeletonInstance, values: dict[str, t.Any]) -> bool:
        return True

    def onEntryUnchanged(self, skel: SkeletonInstance, values: dict[str, t.Any]) -> bool:
        return True

    def onImportFinished(self, moduleName: str, total: int, updated: int, **kwargs: t.Any) -> None:
        pass

    def _kickoff_follow(
        self,
        importdate: dt,
        inform: str | None,
        import_conf_name: str = "import_conf",
        **kwargs: t.Any,
    ) -> None:
        # Check if tree type and nodes where imported, then import the leafs first
        if (
            self.get_handler() == "tree"
            and kwargs.get("skelType") == "node"
            and kwargs.get("_autoSkelType")
        ):
            kwargs["skelType"] = "leaf"
            kwargs["_autoSkelType"] = True
            self.do_import(importdate, inform, follow=True, **kwargs)
            return

        import_conf = getattr(self, import_conf_name)
        import_conf = import_conf() if callable(import_conf) else import_conf

        for name in import_conf.get("follow") or []:
            # mod = getattr(conf.main_app, name, None)
            mod = getattr(conf.main_app.vi, name, None)
            if mod and isinstance(mod, Importable):
                logger.info("%s: Kicking off import for %r", self.moduleName, name)
                mod.do_import(importdate, inform=inform, follow=True, _queue="import")
            else:
                logger.warning("Cannot follow module '%r'", name)

    @CallDeferred
    def do_clear(
        self,
        importdate: dt,
        inform: str | None,
        total: int,
        updated: int,
        import_conf_name: str = "import_conf",
        cursor: str | None = None,
        removed: int = 0,
        follow: bool = False,
        dry_run: bool = False,
        delete_filter: dict | None = None,
        **kwargs: t.Any,
    ) -> None:
        logger.info("do_clear")
        import_conf = getattr(self, import_conf_name)
        import_conf = import_conf() if callable(import_conf) else import_conf

        assert import_conf.get("source"), "No source specified to import from"

        # Mark this request as an import task
        current.request.get().kwargs["isImportTask"] = True

        # Get skeleton
        skel = self.import_skel(skelType=kwargs.get("skelType"))
        assert skel

        q = skel.all()

        if not delete_filter:
            delete_filter = {}

        if conf_filter := import_conf.get("filter"):
            delete_filter.update(conf_filter)

        q.filter("importdate <", importdate)
        q = q.mergeExternalFilter(delete_filter)

        if not q:
            logger.error("filter prohibits clearing")
            return

        if cursor:
            q.setCursor(cursor)

        fetched = 0
        for skel in q.fetch(limit=99):
            logger.debug(
                "%s: Deleting %r with importdate %r",
                self.moduleName,
                skel["key"],
                skel["importdate"],
            )

            if "import_behavior" in skel and skel["import_behavior"] in [
                "only_override",
                "keep",
            ]:
                continue  # don't remove this marked entries

            if "outdated" in skel:
                skel["outdated"] = True

                if dry_run:
                    logger.info(f"dry run outdating {skel=}")
                    continue

                skel.toDB()
            else:
                if dry_run:
                    logger.info(f"dry run deleting {skel=}")
                    continue

                skel.delete()

                if self.get_handler() in ["hierarchy", "tree"]:
                    self.onDeleted(kwargs["skelType"], skel)
                else:
                    self.onDeleted(skel)

            fetched += 1
            removed += 1

        logger.info(
            "%s: %d entries %s",
            self.moduleName,
            fetched,
            "flagged outdated" if "outdated" in skel else "deleted",
        )

        if fetched and (cursor := q.getCursor()):
            self.do_clear(
                importdate,
                inform,
                total,
                updated,
                cursor=cursor,
                removed=removed,
                import_conf_name=import_conf_name,
                follow=follow,
                delete_filter=delete_filter,
                _queue="import",
                **kwargs,
            )
            return

        logger.info(
            "%s: Import finished, %d entries in total, %d updated, %d deleted",
            self.moduleName,
            total,
            updated,
            removed,
        )

        if inform:
            email.sendEMail(
                dests=inform,
                stringTemplate=JINJA_EMAIL_TEMPLATE,
                skel={
                    "sourceportal": import_conf["source"]["url"],
                    "targetportal": conf.instance.project_id,
                    "sourcemodule": import_conf.get("module", self.moduleName),
                    "targetmodule": self.moduleName,
                    "total": total,
                    "updated": updated,
                    "removed": removed,
                },
            )

        self.onImportFinished(self.moduleName, total, updated, **kwargs)

        if follow:
            self._kickoff_follow(importdate, inform, **kwargs)

    # types from viur-core prototypes

    moduleName: str
    onAdded: types.MethodType
    onEdited: types.MethodType
    onDeleted: types.MethodType
    describe: types.MethodType
    editSkel: types.MethodType



import logging as _logging

from .callable_tasks import *
from .checks import *
from .context import *
from .db import *
from .decorators import *
from .helpers import *
from .importer import *
from .memcache import *
from .numeric import *
from .property import *
from .report import *
from .viur import *

# By default, the toolkit log level is INFO
if not _logging.getLogger(__name__).level:
    _logging.getLogger(__name__).setLevel(_logging.INFO)
del _logging



import logging
import time
import typing as t

from viur.core import bones, db, skeleton

__all__ = [
    "normalize_key",
    "write_in_transaction",
    "increase_counter",
    "set_status",
]

logger = logging.getLogger(__name__)

_KeyType: t.TypeAlias = str | db.Key


def normalize_key(key: _KeyType) -> db.Key:
    if isinstance(key, str):
        return db.Key.from_legacy_urlsafe(key)
    elif isinstance(key, db.Key):
        return key
    raise TypeError(f"Expected key of type str or db.Key, got: {type(key)}")


def write_in_transaction(key: _KeyType, create_missing_entity: bool = True, **values: t.Any) -> db.Entity:
    def txn(_key: db.Key, _values: dict[str, t.Any]) -> db.Entity:
        try:
            entity = db.Get(_key)
        except db.NotFoundError:
            if create_missing_entity:
                entity = db.Entity(_key)
            else:
                raise
        for k, v in _values.items():
            entity[k] = v
        db.Put(entity)
        return entity

    return db.RunInTransaction(txn, normalize_key(key), values)


def increase_counter(key: _KeyType, name: str, value: float | int = 1, start: float | int = 0) -> int | float:
    def txn(_key: db.Key, _name: str, _value: float | int, _start: float | int) -> float | int:
        try:
            entity = db.Get(_key)
        except db.NotFoundError:
            # Use not db.GetOrInsert here, we write the entity later anyway
            # and can therefore save the db.Put in db.GetOrInsert
            entity = db.Entity(_key)

        if _name not in entity:
            entity[_name] = _start
        old_value = entity[_name]
        entity[_name] += _value
        db.Put(entity)
        return old_value

    return db.RunInTransaction(txn, normalize_key(key), name, value, start)


def set_status(
    key: _KeyType,
    values: dict | t.Callable[[skeleton.SkeletonInstance | db.Entity], None],
    precondition: t.Optional[dict | t.Callable[[skeleton.SkeletonInstance | db.Entity], None]] = None,
    create: dict[str, t.Any] | t.Callable[[skeleton.SkeletonInstance | db.Entity], None] | bool = False,
    skel: t.Optional[skeleton.SkeletonInstance] = None,
    update_relations: bool = False,
    retry: int = 1,
) -> skeleton.SkeletonInstance | db.Entity:
    """
    Universal function to set values of an entity within a transaction.
    It is mostly used for status changes, but can also change any value.

    :param key: Entity key to change
    :param values: A dict of key-values to update on the entry, or a callable that is executed within the transaction
    :param precondition: An optional dict of key-values to check on the entry before; can also be a callable.
    :param create: When key does not exist, create it, optionally with values from provided dict, or in a callable.
    :param skel: Use assigned skeleton instead of low-level DB-API
    :param update_relations: Trigger update relations task on success (only in skel-mode, defaults to False)
    :param retry: On ViurDatastoreError, retry for this amount of times.

    If the function does not raise an Exception, all went well.
    It returns either the assigned skel, or the db.Entity on success.
    """

    # Transactional function
    def transaction() -> skeleton.SkeletonInstance | db.Entity:
        exists = True

        # Use skel or db.Entity
        if skel:
            if not skel.fromDB(key):
                if not create:
                    raise ValueError(f"Entity {key=} not found")

                skel["key"] = key
                exists = False

            obj = skel
        else:
            obj = db.Get(key)

            if obj is None:
                if not create:
                    raise ValueError(f"Entity {key=} not found")

                obj = db.Entity(key)
                exists = False

        # Handle create
        if not exists and create:
            if isinstance(create, dict):
                for bone, value in create.items():
                    obj[bone] = value
            elif callable(create):
                create(obj)

        # Handle precondition
        if isinstance(precondition, dict):
            for bone, value in precondition.items():
                if obj[bone] != value:
                    raise ValueError(f"{bone} contains {obj[bone]!r}, expecting {value!r}")

        elif callable(precondition):
            precondition(obj)

        # Set values
        if isinstance(values, dict):
            for bone, value in values.items():
                # Increment by value?
                if bone[0] == "+":
                    obj[bone[1:]] += value
                # Decrement by value?
                elif bone[0] == "-":
                    obj[bone[1:]] -= value
                else:
                    if skel and (
                        (boneinst := getattr(skel, bone, None))
                        and isinstance(boneinst, bones.RelationalBone)
                    ):
                        assert skel.setBoneValue(bone, value)
                        continue

                    obj[bone] = value

        elif callable(values):
            values(obj)

        else:
            raise ValueError("'values' must either be dict or callable.")

        if skel:
            assert skel.toDB(update_relations=update_relations)
        else:
            db.Put(obj)

        return obj

    # Retry loop
    while True:
        try:
            return db.RunInTransaction(transaction)

        except db.ViurDatastoreError as e:
            retry -= 1
            if retry <= 0:
                raise

            logging.debug(f"{e}, retrying {retry} more times")

        time.sleep(1)



import functools
import json
import logging
import types
import typing as t

import deprecated

from viur.core import current, errors

__all__ = [
    "as_json_response",
    "cache_call_for_request",
    "debug",
    "parse_request_payload",
]

logger = logging.getLogger(__name__)

P = t.ParamSpec("P")
T = t.TypeVar("T")


def debug(func: t.Callable[P, T]) -> t.Callable[P, T]:
    """Decorator to print the function signature and return value"""

    @functools.wraps(func)
    def wrapper_debug(*args: P.args, **kwargs: P.kwargs) -> T:
        args_repr = list(map(repr, args))
        kwargs_repr = [f"{k!s}={v!r}" for k, v in kwargs.items()]
        signature = ", ".join(args_repr + kwargs_repr)
        logging.info(f"CALLING {func.__name__}({signature})")
        value = func(*args, **kwargs)
        logging.info(f"{func.__name__} RETURNED {value}")
        return value

    return wrapper_debug


def as_json_response(
    func: t.Optional[t.Callable[P, T]] = None,
    **decorator_kw_args: t.Any,
) -> t.Callable[P, str] | t.Callable[[t.Callable[P, T]], t.Callable[P, str]]:
    """Decorator that returns the method/function response json serialized.

    Optional callable to pass keyword-arguments for the json.dumps-call.

    Example:
        >>> from viur.core import exposed
        >>> @exposed
        >>> @as_json_response(default=str)
        >>> def your_method(self):
        >>>     return {"foo": "bar"}
    """

    def outer_wrapper(f: t.Callable[P, T]) -> t.Callable[P, str]:
        @functools.wraps(f)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> str:
            current.request.get().response.headers["Content-Type"] = "application/json"
            return json.dumps(f(*args, **kwargs), **decorator_kw_args)

        return wrapper

    if isinstance(func, (types.MethodType, types.FunctionType)):
        return outer_wrapper(func)  # @as_json_response
    else:
        return outer_wrapper  # @as_json_response() or @as_json_response(**any_kwargs)


asJsonResponse = deprecated.deprecated(as_json_response)


class ParseFunc(t.Protocol):
    def __call__(self, content_type: str, payload: str) -> t.Any:
        ...


def parse_request_payload(
    func: t.Optional[t.Callable[P, T]] = None,
    accept_only: None | str | list[str] | set[str] | tuple[str] = None,
    parse_func: None | ParseFunc = None,
) -> t.Callable[P, T] | t.Callable[[t.Callable[P, T]], t.Callable[P, T]]:
    """Decorator that parses the payload of the request.

    The payload will be parsed based on the provided content-type header
    by the set parseFunc.
    The wrapped function MUST have the argument *payload* where the
    parsed payload/payload will be passed through.

    Example:
        >>> from viur.core import exposed
        >>> @exposed
        >>> @parse_request_payload(accept_only="application/json")
        >>> def your_method(self, payload):
        >>>     # do something with the payload - it's a python object

    :param func: The wrapped function/method when not called
    :param accept_only: The accepted content-types for this ressource.
        As specified in https://tools.ietf.org/html/rfc2045#section-5
        Set to None to accept all (*/*).
    :param parse_func: a custom function that parse the payload

    :return: The result of the wrapped function
    """

    def default_parse_func(content_type: str, payload: str) -> t.Any:
        if content_type == "application/json":
            return json.loads(payload)
        raise NotImplementedError(f"No parser for type '{content_type}' implemented")

    def outer_wrapper(f: t.Callable[P, T]) -> t.Callable[P, T]:
        @functools.wraps(f)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            cur_req = current.request.get()
            # Set the Accept-header
            if accept_only is None:
                cur_req.response.headers["Accept"] = "*/*"
            else:
                cur_req.response.headers["Accept"] = ";".join(accept_only)
            # Check the given content-type
            content_type = cur_req.request.headers["Content-Type"].split(";", 1)[0].lower()
            if accept_only is not None and content_type not in accept_only:
                logger.error("Unsupported format '%s' for payload", content_type)
                raise errors.HTTPException(
                    415, "Unsupported Media Type",
                    "Unsupported format '%s' for payload" % content_type
                )
            # Parse the payload
            try:
                payload = parse_func(content_type, cur_req.request.body)  # type: ignore
            except ValueError:  # Thrown by the parser to indicate that the payload is unparsable
                logger.exception("Failed to parse payload of type '%s'", content_type)
                raise errors.HTTPException(
                    400, "Bad Request",
                    "Failed to parse the payload"
                )
            except NotImplementedError:
                logger.exception("Parser does not support the given type '%s'", content_type)
                raise errors.NotImplemented(
                    "Parser does not support the given type '%s'" % content_type
                )
            # Call the wrapped func with the additionally payload-argument
            return f(*args, payload=payload, **kwargs)

        return wrapper

    # Check arguments first
    if isinstance(accept_only, str):
        accept_only = [accept_only]
    elif not (accept_only is None or isinstance(accept_only, (list, tuple, set))):
        raise TypeError("Argument accept_only must be a str or iterable of str.")
    if parse_func is None:
        parse_func = default_parse_func
    elif not isinstance(parse_func, (types.MethodType, types.FunctionType)):
        raise TypeError("Invalid parse_func given")

    # Check whether the decorator was called or not
    if isinstance(func, (types.MethodType, types.FunctionType)):
        return outer_wrapper(func)  # @parse_request_payload
    else:
        return outer_wrapper  # @parse_request_payload() or @parse_request_payload(**anyKwargs)


parseRequestPayload = deprecated.deprecated(parse_request_payload)


def cache_call_for_request(
    func: t.Optional[t.Callable[P, T]] = None,
) -> t.Callable[P, T] | t.Callable[[t.Callable[P, T]], t.Callable[P, T]]:
    """Cache method calls for the current request

    args, kwargs must be hashhable.
    """

    # TODO: documentation, merge default kwargs into args
    def outer_wrapper(f: t.Callable[P, T]) -> t.Callable[P, T]:
        @functools.wraps(f)
        def inner_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            assert func is not None
            cache_key = tuple([func.__name__] + list(args) + list(sorted(kwargs.items())))
            cache = current.request_data.get().setdefault("_call_cache", {})

            try:
                res = cache[cache_key]
                logger.debug(f"Loaded {cache_key} from cache (req cache hit)")
                return res
            except KeyError:
                res = cache[cache_key] = f(*args, **kwargs)
                logger.debug(f"Stored {cache_key} in cache (req cache miss)")
                return res

        return inner_wrapper

    if isinstance(func, (types.MethodType, types.FunctionType)):
        return outer_wrapper(func)  # @cache_call_for_request
    else:
        return outer_wrapper  # @cache_call_for_request() or @cache_call_for_request(**any_kwargs)



import csv
import io
import json
import logging
import typing as t
from datetime import datetime as dt
from types import TracebackType

from google.cloud.storage import Bucket

from viur.core import email
from viur.core.modules.file import GOOGLE_STORAGE_BUCKET

__all__ = ["Report"]

logger = logging.getLogger(__name__)

GOOGLE_STORAGE_BUCKET: Bucket = GOOGLE_STORAGE_BUCKET
"""Main GOOGLE_STORAGE_BUCKET (here reassigned to add the type hint)"""


class Report:
    """Reports are a kind of logging

    For example tasks in several there are information one want to save and
    finally send as mail to a person.

    .. code-block:: python

        from viur import toolkit

        with toolkit.Report("my_task_report") as report:
            # ... do something ...
            report.write(**{  # write some results
                "action": "update",
                "skel_key": "bar",
                "active": True,
            })
            # Not every entry must contain the same keys
            report.write(action="add", active=False)

        # In the last task batch send the mail
        with toolkit.Report("my_task_report") as report:
            report.send_as_mail("mail@me.com")
            report.delete()
    """

    __slots__ = ("name", "content", "columns", "changed")

    def __init__(self, name: str):
        self.name: str = name
        self.content: list[dict[str, t.Any]] = []
        self.columns: set = set()
        self.changed: bool = False

    @property
    def gsc_name(self) -> str:
        return "reports/report_{0}.json".format(self.name)

    def _read(self, raise_exc: bool = False) -> None:
        if (blob := GOOGLE_STORAGE_BUCKET.get_blob(self.gsc_name)) is None:
            if raise_exc:
                raise ValueError(f"GSC blob {self.gsc_name} not found")
            return None
        columns, content = json.loads(
            blob.download_as_bytes().decode("utf-8")
        )
        self.content = content
        self.columns = set(columns)

    def write(self, **value: t.Any) -> None:
        value["reportdate"] = dt.now().isoformat(" ", "seconds")
        self.content.append(value)
        self.columns.update(value.keys())
        self.changed = True

    def flush(self) -> None:
        if not self.changed:
            return
        value = (list(self.columns), self.content[:])
        GOOGLE_STORAGE_BUCKET.blob(self.gsc_name).upload_from_file(
            file_obj=io.BytesIO(json.dumps(value).encode()),
            content_type="application/json"
        )
        self.changed = False

    def read(self) -> list[dict[str, t.Any]]:
        self._read()
        return self.content

    def delete(self) -> None:
        # mark as not changed, so it won't be flushed
        self.changed = False
        try:
            GOOGLE_STORAGE_BUCKET.blob(self.gsc_name).delete()
        except AttributeError:  # already deleted
            logger.debug(self.gsc_name, exc_info=True)

    def send_as_mail(self, receiver: str | list[str]) -> bool:
        header = list(self.columns)
        body = [
            {key: entry.get(key) for key in header}
            for entry in self.content
        ]

        sio = io.StringIO()
        if self.content:
            writer = csv.DictWriter(sio, fieldnames=header)
            writer.writerows(self.content)
        else:
            sio.write("- no entries -")

        # noinspection PyTypeChecker
        return email.sendEMail(
            dests=receiver,
            # language=Jinja2
            stringTemplate="""Report {{ skel["name"] }}
<table>
    <thead>
        <tr>
            {% for col in skel["header"] %}
                <th>{{ col }}</th>
            {% endfor %}
        </tr>
    </thead>
    <tbody>
        {% if not skel["body"] %}
            <tr>- no entries -</td></tr>
        {% endif %}
        {% for row in skel["body"] %}
            {%- set style = "" -%}
            {%- set level = (row["level"] or "").lower() -%}
            {%- if level == "info" -%}
                {%- set style = "color:dodgerblue" -%}
            {%- elif level == "warning" -%}
                {%- set style = "color:orange" -%}
            {%- elif level == "error" -%}
                {%- set style = "color:red" -%}
            {%- endif -%}
            <tr style="{{ style }}">
                {% for col in row.values() %}
                    <td>{{ col }}</td>
                {% endfor %}
            </tr>
        {% endfor %}
    </tbody>
</table>""",
            skel={
                "name": self.name,
                "header": header,
                "body": body,
            },
            attachments=[{
                "filename": "report.csv",
                "mimetype": "text/csv",
                "content": sio.getvalue().encode(),
            }],
        )

    def __enter__(self) -> t.Self:
        self._read()
        return self

    def __exit__(
        self,
        exc_type: t.Optional[t.Type[BaseException]],
        exc_val: t.Optional[BaseException],
        exc_tb: t.Optional[TracebackType],
    ) -> t.Literal[False]:
        self.flush()
        return False



# This is the place where the viur-toolkit version number is defined;
# For pre-releases, postfix with ".devN", ".betaN" or ".rcN" where `N` is an incremented number for each pre-release.
# This will mark it as a pre-release as well on PyPI.
# See CONTRIBUTING.md for further information. # TODO: tbd

__version__ = "0.1.4"

assert __version__.count(".") >= 2 and "".join(__version__.split(".", 3)[:3]).isdigit(), \
    "Semantic __version__ expected!"



"""
`Importer` is a `requests.Session`-inherited class which provides helpers to convert
from JSON into a skeleton structure.
"""

import html
import json
import logging
import mimetypes
import numbers
import re
import typing as t

import requests
from requests import Response

from viur.core import bones, conf, db, skeleton, utils

if t.TYPE_CHECKING:
    from .importable import Importable

NIL: t.Final = object()

logger = logging.getLogger(__name__)


class Importer(requests.Session):

    def __init__(self, source: dict, render: str = "vi", cookies: dict | None = None):
        from .. import get_task_retry_count

        super().__init__()

        url = source["url"]
        if not url:
            raise IOError("Importer disabled by configuration")

        self.host = url
        self.render = render
        self.method = source.get("auth")
        self.secretname = self.user = self.password = self.key = self.otp = None

        if cookies:
            logger.info(f"Using existing session cookies {cookies.keys()}")
            self.cookies.update(cookies)

            if get_task_retry_count() > 1 and self.get("/user/view/self").status_code == 401:
                logger.warning("Got 401, session seems to be expired, Re-Login")
                self.cookies.clear()
            else:
                return

        if self.method == "userpassword":
            self.user = source["user"]
            self.password = source["pass"]
        elif self.method == "userpassword+otp":
            self.user = source["user"]
            self.password = source["pass"]
            self.otp = source["otp"]
        elif self.method in ["secretkey", "loginkey"]:
            self.secretname = source.get("keyname", "secret")

            if module := source.get("from"):
                assert "." in module
                module, key = module.split(".", 1)

                mod = getattr(conf.main_app.vi, module)  # fixme: viur-core 3.7 Remove vi when possible!
                mod = db.Get(db.Key(mod.viewSkel().kindName, mod.getKey()))
                self.key = mod[key]
            else:
                self.key = source.get("key")

            assert self.key, "No key defined?"

        elif self.method is not None:
            raise ValueError(f"Unsupported authmethod configuration {self.method!r}")

        if self.method and not self.login():
            raise IOError(f"Unable to logon to '{self.host}'")

    def logout(self) -> t.Any:
        if self.method and self.method != "secretkey":
            return self.get("/user/logout", params={"skey": self.skey()}, timeout=30).json()

    def skey(self) -> str:
        return self.get("/skey", timeout=30).json()

    def get(self, url: str, *args: t.Any, **kwargs: t.Any) -> Response:  # type: ignore[override]
        if url.startswith("/"):
            url = url[1:]

        if self.method == "secretkey":
            if "data" not in kwargs:
                kwargs["data"] = {}

            kwargs["data"][self.secretname] = self.key
        else:
            logger.debug(f"GET  {'/'.join([self.host, self.render, url])} {kwargs}")

        return super().get("/".join([self.host, self.render, url]), *args, **kwargs)

    def post(self, url: str, *args: t.Any, **kwargs: t.Any) -> Response:  # type: ignore[override]
        if url.startswith("/"):
            url = url[1:]

        if self.method == "secretkey":
            if "data" not in kwargs:
                kwargs["data"] = {}

            kwargs["data"][self.secretname] = self.key
        else:
            logger.debug(f"POST {'/'.join([self.host, self.render, url])} {kwargs}")

        return super().post("/".join([self.host, self.render, url]), *args, **kwargs)

    def login(self) -> bool:
        if self.method == "secretkey":
            logger.debug(f"{self.method=} requires no login")
        else:
            if self.method in ("userpassword", "userpassword+otp"):
                answ = self.post("/user/auth_userpassword/login",
                                 data={"name": self.user,
                                       "password": self.password,
                                       "skey": self.skey()},
                                 timeout=30)
            elif self.method == "loginkey":
                answ = self.post("/user/auth_loginkey/login",
                                 data={"key": self.key,
                                       "skey": self.skey()},
                                 timeout=30)

            while True:
                if answ.status_code == 200:
                    try:
                        res = answ.json()
                    except (json.decoder.JSONDecodeError, requests.exceptions.JSONDecodeError):
                        try:
                            res = json.loads(
                                re.search(r"JSON\(\((.*)\)\)", answ.text).group(1))  # type:ignore[union-attr]
                        except Exception:
                            res = None

                    if res:
                        if res == "OKAY":
                            break
                        elif self.method == "userpassword+otp" and res["action"] == "otp":
                            answ = self.post("/user/f2_timebasedotp/otp",
                                             data={"otptoken": self.otp,
                                                   "skey": self.skey()},
                                             timeout=30)
                            continue

                logger.error(f"Unable to logon to {self.host!r}, got {answ=} with {answ.text=}")
                return False

        logger.debug(f"HELLO {self.host}")
        return True

    def list(self, module: str, **kwargs: t.Any) -> dict[str, t.Any] | None:
        req = self.get(f"/{module}/list", params=kwargs, timeout=60)

        if not req.status_code == 200:
            logger.error(f"Error {req.status_code!r}, unable to fetch items")
            return None

        return req.json()

    def flatlist(self, module: str, **kwargs: t.Any) -> dict[str, t.Any] | None:
        req = self.get(f"/{module}/listentries", params=kwargs, timeout=60)

        if not req.status_code == 200:
            logger.error(f"Error {req.status_code!r}, unable to fetch items")
            return None

        return req.json()

    def flatten_relational_data(self, bone_name: str, data: t.Any) -> dict[str, t.Any]:
        ret = {}
        if isinstance(data, list) or isinstance(data, dict):
            if isinstance(data, dict):
                data = [data]
            for i, entry in enumerate(data):
                prefix = bone_name + ".%d." % i
                ret[prefix + "key"] = entry["dest"]["key"]
                if entry["dest"]["rel"]:
                    for key, value in entry["dest"]["rel"].items():
                        ret[prefix + key] = value
        else:
            ret[bone_name] = data
        return ret

    def view(self, module: str, key: str) -> dict[str, t.Any] | None:
        req = self.get(f"/{module}/view/{key}", timeout=10)

        if not req.status_code == 200:
            logger.error(f"Error {req.status_code!r}, unable to fetch entry")
            return None

        return req.json()

    def import_file(self, info: dict[str, t.Any]) -> None | db.Key:
        assert "dlkey" in info and "name" in info

        name = info["name"]
        dlkey = info["dlkey"]
        servingurl = info.get("servingurl")
        try:
            size = info["size"]
        except KeyError:
            logger.warning(f"size is missing in {info=}")
            size = -1

        if not isinstance(size, int):
            try:
                size = int(size)
            except ValueError:
                logger.error("cannot convert {size!r} to int", size)
                return None

        if size > 200_000_000:
            logger.error("Download skipped, because its larger than 200mb")
            logger.error(info)
            return None

        file_skel_cls = skeleton.skeletonByKind("file")
        if "import_key" in dir(file_skel_cls):
            logger.debug(f'[importFile] Looking for {info["key"]}')

            if file_skel := file_skel_cls().all().filter("import_key =", info["key"]).getSkel():
                logger.debug(f"[importFile] File {info} already imported as {file_skel}")
                return file_skel["key"]

            logger.debug(f"[importFile] File {info} not imported yet")

        if not (mimetype := info.get("mimetype")):
            mimetype = mimetypes.guess_type(name)[0]

        if "downloadUrl" in info:
            # ViUR 3
            logger.debug(f"Downloading {self.host + info['downloadUrl']}")
            res = self.get(info["downloadUrl"], timeout=60)
        else:
            # ViUR 2
            logger.debug(f"Downloading {self.host + '/file/download/' + dlkey}")
            res = self.get("/file/download/" + dlkey, timeout=60)
        if not res.ok:
            if servingurl:
                logger.warning(f"Download failed, trying to fetch using {servingurl=}")
                content = super().get(servingurl).content
            else:
                logger.error(f"Download failed with {res.status_code=}")
                return None
        else:
            content = res.content

        logger.debug(f"{name=} has {len(content)!r} bytes")

        if "import_key" in dir(file_skel_cls):
            return conf.main_app.vi.file.write(name, content, mimetype, import_key=info["key"])

        return conf.main_app.file.write(name, content, mimetype)

    def set_skel_value(self, skel: skeleton.SkeletonInstance, bone_name: str, value: t.Any, debug: bool = False) -> int:
        # FIXME: This method is a total mess up of bone types and nested structure.
        #        It needs a complete refactoring (split into methods), should use more bone methods
        #        and a more robust behaviour against invalid bone data

        changes = 0
        bone = getattr(skel, bone_name)

        if bone is None:
            return changes

        bone_value = skel[bone_name]
        if isinstance(bone, bones.FileBone):
            knownFiles = {}

            if debug:
                logger.debug(f"{bone_value=}")

            if bone_value:
                if bone.languages and isinstance(value, dict):
                    if bone_value is None:
                        bone_value = {}

                    for lang, val in value.items():
                        if lang in bone.languages:
                            if not bone_value.get(lang):
                                continue

                            if isinstance(bone_value[lang], dict):
                                knownFiles[bone_value[lang]["dest"]["name"]] = bone_value[lang]
                            else:
                                for entry in bone_value[lang]:
                                    knownFiles[entry["dest"]["name"]] = entry

                elif isinstance(skel[bone_name], dict):
                    knownFiles[skel[bone_name]["dest"]["name"]] = skel[bone_name]
                else:
                    for entry in skel[bone_name]:
                        knownFiles[entry["dest"]["name"]] = entry

            # knownFiles.clear() #Temporary enabled to clear all files

            if debug:
                logger.debug(f"{bone_name} knownFiles = {knownFiles!r}")

            def handle_entries(changes: int, value: t.Any, lang: str | None = None) -> int:
                for entry in value:
                    # fixme: not sure why, but while importing fluidpagecontent files, some entries where None... skipping
                    if not entry:
                        continue

                    if debug:
                        logger.debug(f"{entry=} // {lang=}")

                    fileName = entry["dest"]["name"]
                    key = None

                    if fileName not in knownFiles:
                        if debug:
                            logger.debug(f"{bone_name} name {fileName=} is not known")

                        key = self.import_file(entry["dest"])
                        if not key:
                            continue
                        # assert (key := self.import_file(entry["dest"]))

                    if not bone.using:
                        if key:
                            if skel.setBoneValue(bone_name, key, append=bone.multiple, language=lang):
                                changes += 1
                            else:
                                logger.error(f"Unable to set bone {bone_name} to {key}")

                    else:
                        assert entry["rel"]
                        using_skel = bone.using()
                        using_skel.unserialize(knownFiles[fileName]["rel"] if fileName in knownFiles else {})
                        changes += self.translate(using_skel, entry["rel"])

                        if key:
                            if skel.setBoneValue(bone_name, (key, using_skel), append=bone.multiple, language=lang):
                                changes += 1
                            else:
                                logger.error(f"Unable to set bone {bone_name} to {key}")
                        else:
                            knownFiles[fileName]["rel"].unserialize(using_skel.dbEntity)
                return changes

            if bone.languages and isinstance(value, dict):
                for lang, val in value.items():
                    if lang in bone.languages:
                        if not bone_value.get(lang):
                            continue

                        if isinstance(val, dict):
                            val = [val]
                        changes = handle_entries(changes, val, lang)

            else:

                if isinstance(value, dict):
                    value = [value]

                changes = handle_entries(changes, value)

        elif isinstance(bone, bones.RelationalBone):

            if isinstance(value, (dict, str, type(db.Key))):
                value = [value]
            elif value is None:
                value = []

            if bone_value and isinstance(bone_value, (dict, str, type(db.Key))):
                bone_value = [bone_value]
            elif not bone_value:
                bone_value = []

            skel[bone_name] = None

            # A difference in length is always a diff indicator!
            if len(value) != len(bone_value):
                changes += 1

            # Deep iterate through bone values
            for nentry in value:
                if bone_value:
                    # There is an old entry
                    oentry = bone_value.pop(0)

                else:
                    # Old entry does not exist
                    oentry = None

                if bone.using:
                    using_skel = bone.using()
                else:
                    using_skel = None

                # Rewrite the key of the referenced entity
                if isinstance(nentry, dict):
                    assert "rel" in nentry and "dest" in nentry, "This doesn't look like a RelationalBone"
                    key = db.Key(bone.kind, db.KeyClass.from_legacy_urlsafe(nentry["dest"]["key"]).id_or_name)

                elif isinstance(nentry, (str, db.Key)):
                    key = nentry
                else:
                    raise ValueError()

                # In case of an using_skel, check for differences also.
                if using_skel:
                    using_skel.unserialize(oentry["rel"] if oentry else {})
                    changes += self.translate(
                        using_skel, nentry["rel"] or {},
                        {k: k for k in using_skel.keys()},
                        debug=debug
                    )

                    if not skel.setBoneValue(bone_name, (key, using_skel), append=bone.multiple):
                        logger.error(f"Unable to set bone {bone_name} to {key}")

                elif not skel.setBoneValue(bone_name, key, append=bone.multiple):
                    logger.error(f"Unable to set bone {bone_name} to {key}")

        elif isinstance(bone, bones.RecordBone):
            def set_value(val: t.Any, lang: str | None) -> None:
                nonlocal bone_value, changes, key

                if isinstance(val, (dict, str, db.Key)):
                    val = [val]
                elif val is None:
                    val = []

                if bone_value and isinstance(bone_value, (dict, str, db.Key)):
                    bone_value = [bone_value]
                elif not bone_value:
                    bone_value = []

                if bone.languages:
                    skel[bone_name] = {lang: [] if bone.multiple else None
                                       for lang in bone.languages}
                else:
                    skel[bone_name] = [] if bone.multiple else None

                for nentry in val:
                    if bone_value:
                        # There is an old entry
                        oentry = bone_value.pop(0)

                    else:
                        # Old entry does not exist
                        oentry = None
                        changes += 1

                    using_skel = bone.using()
                    using_skel.unserialize(oentry or {})

                    changes += self.translate(using_skel, nentry, {k: k for k in using_skel.keys()}, debug=debug)

                    if debug:
                        logger.debug(f"Assign {bone_name=}, {using_skel=}, append={bone.multiple}, language={lang}")
                    if not skel.setBoneValue(bone_name, using_skel, append=bone.multiple, language=lang):
                        logger.error(f"Unable to set bone {bone_name}.{lang} to {using_skel}")

            if debug:
                logger.debug(f"{bone.languages=} // {value}")

            if bone.languages:
                for lang in bone.languages:
                    if value.get(lang):
                        set_value(value[lang], lang)
            else:
                set_value(value, None)

        elif bone.languages and isinstance(value, dict):
            if bone_value is None:
                bone_value = {}

            if debug:
                logger.debug(f"{bone_value=} {value=} {bone=}")

            for lang, val in value.items():
                if lang in bone.languages:
                    if lang not in bone_value:
                        changes += 1
                    elif bone_value[lang] != val:
                        changes += 1

                    if bone.multiple:
                        if val is None:
                            val = []
                        elif isinstance(val, (str, numbers.Number)):
                            val = [val]
                        elif not isinstance(val, list):
                            logger.warning(f"Unexpected {val=}")

                    logger.debug(f"{val=} IN")

                    if isinstance(bone, bones.BooleanBone):
                        val = [utils.parse.bool(value) for value in val]

                    elif isinstance(bone, bones.NumericBone):
                        values = []
                        for value in val:
                            try:
                                value = float(value)

                            except ValueError:
                                logger.error(f"Not a numeric value for {bone_name}: {value=}")
                                value = bone.getDefaultValue()

                            values.append(value)

                        val = values

                    logger.debug(f"{val=} OUT")

                    try:
                        if not skel.setBoneValue(bone_name, val, language=lang):
                            logger.error(f"Failed to set {bone_name=} {lang=} to {value=}")

                    except Exception as exc:  # noqa
                        exc.add_note(f"{bone_name=} | {bone_value=} | {lang=} | {val=}")
                        raise

        elif isinstance(bone, bones.DateBone):
            skel.setBoneValue(bone_name, value)

            if bone_value != skel[bone_name]:
                changes += 1

        else:  # TODO: this is duplicate code (it's just the inner part of the language for loop)
            if bone.multiple:
                if not isinstance(value, list):
                    value = [value]
            else:
                if isinstance(value, list) and len(value) == 1:
                    value = value[0]

            if isinstance(bone, bones.BooleanBone):
                logger.debug(f"{value=} [IN]")

                if bone.multiple:
                    value = [utils.parse.bool(v) for v in value]
                else:
                    value = utils.parse.bool(value)

                logger.debug(f"{value=} [OUT]")

            new_value_as_str = html.unescape(str(bone_value)).strip()
            old_value_as_str = html.unescape(str(value)).strip()

            if debug:
                logger.debug(
                    f"{bone_name} old={old_value_as_str!r} != new:{new_value_as_str!r}? "
                    f"{new_value_as_str != old_value_as_str!r}"
                )

            if new_value_as_str != old_value_as_str:
                if bone.multiple and not isinstance(value, list):
                    value = [value]

                if not skel.setBoneValue(bone_name, value):
                    logger.error(f"Failed to set {bone_name=} to {value=}")

                changes += 1

                if debug:
                    logger.debug(f"{bone_name} new value {skel[bone_name]=}")

        return changes

    def translate(
        self,
        skel: skeleton.SkeletonInstance,
        values: dict[str, t.Any],
        translate: dict[str, str | t.Callable] | None = None,
        debug: bool = False,
    ) -> int:
        changes = 0

        if translate is None:
            translate = {k: k for k in skel.keys()}

        for bone, tr in translate.items():
            if debug:
                logger.debug(f"{bone=}, {tr=}, {values.get(bone)=}")

            if bone.endswith("*"):
                for k, v in values.items():
                    if k.startswith(bone[:-1]):
                        tail = k[len(bone[:-1]):]

                        if isinstance(tr, str):
                            if tr:
                                if tr.endswith("*"):
                                    lookfor = tr[:-1] + tail
                                else:
                                    lookfor = tr + tail
                            else:
                                lookfor = k

                            if lookfor in skel:
                                old = skel[lookfor]

                                # Set to empty (don't do this with set_skel_value)
                                if not v:
                                    if skel[lookfor]:
                                        if debug:
                                            logger.debug(f"{lookfor!r} changed from {old!r} to {v!r}")

                                        changes += 1

                                    skel[lookfor] = getattr(skel, lookfor).getDefaultValue(skel)

                                elif ch := self.set_skel_value(skel, lookfor, v, debug=debug):
                                    if debug:
                                        logger.debug(f"{lookfor!r} changed from {old!r} to {skel[lookfor]!r}")

                                    changes += ch

                        elif callable(tr):
                            changes += tr(
                                skel=skel,
                                bone=bone,
                                value=values[k],
                                values=values,
                                module=self
                                # TODO: k is missing (key from incoming values)
                            )

            elif bone in values:

                if tr and isinstance(tr, str):
                    if tr in skel:
                        old = skel[tr]

                        # Set to empty (don't do this with set_skel_value)
                        if not values[bone]:
                            if skel[tr]:
                                if debug:
                                    logger.debug(f"{tr!r} changed from {old!r} to {values[bone]!r}")

                                changes += 1
                            bone_instance = getattr(skel, tr)
                            if bone_instance.multiple:
                                skel[tr] = []
                            else:
                                skel[tr] = bone_instance.getDefaultValue(skel)
                        else:
                            if debug:
                                logger.debug(f"setting {tr!r} to {values[bone]!r}")

                            try:
                                if ch := self.set_skel_value(skel, tr, values[bone], debug=debug):
                                    if debug:
                                        logger.debug(f"{tr!r} changed from {old!r} to {skel[tr]!r}")

                                    changes += ch

                            except:  # db.Error:
                                raise
                                logger.warning(f"Unable to set value for {tr=!r}")

                elif callable(tr):
                    changes += tr(
                        skel=skel,
                        bone=bone,
                        value=values[bone],
                        values=values,
                        module=self
                    )

            elif bone in skel and callable(tr):
                changes += tr(
                    skel=skel,
                    bone=bone,
                    value=NIL,
                    values=values,
                    module=self
                )

            elif debug:
                logger.debug(f"Skipping {bone=}")

        return changes

    def values_to_skel(
        self,
        skel: skeleton.SkeletonInstance,
        values: dict,
        translate: dict = {},
        reset: t.Optional[str | t.Iterable[str]] = None,
        source_key: str = "key",
        update: bool = True,
        enforce: bool = False,
        debug: bool = False,
        module: t.Optional["Importable"] = None,
    ) -> int:
        # assert isinstance(skel, skeleton.BaseSkeleton), "'skel' must be a BaseSkeleton instance"
        assert source_key in values, f"'{source_key}' not in values"

        changes = 0
        exists = True

        try:
            key = db.KeyClass.from_legacy_urlsafe(values[source_key])
        except AttributeError:
            key = db.keyHelper(values[source_key], skel.kindName)

        key = db.Key(skel.kindName, key.id_or_name)

        if module is not None:
            key = module.modify_skel_key(key, skel, values, translate, source_key)

        success = skel.fromDB(key)

        if "import_behavior" in skel and skel["import_behavior"] in ("preserve", "keep"):
            return changes  # dont change skels with this behaviors

        if not enforce and success:
            if not update:
                logger.debug(f"{skel.kindName} entity with {key=} exists, but no update wanted")
                return changes
        else:
            if enforce:
                logger.debug(f"Creation of {skel.kindName} entity with {key=} will be enforced.")
            else:
                logger.debug(f"{skel.kindName} entity with {key=} does not not exist")
                exists = False

            skel["key"] = key
            changes += 1

        if reset:
            if isinstance(reset, list):
                for bone in reset:
                    skel[bone] = None

            else:
                assert isinstance(reset, str)
                skel[reset] = None

        changes += self.translate(skel, values, translate, debug=debug)

        if changes:
            logger.debug(f"{skel.kindName} {key=} detected {changes=}")

        if debug:
            logger.debug("--- values_to_skel ---")

            for bone in skel.keys():
                logger.debug(
                    f"{bone=} (multiple={getattr(skel, bone).multiple}, languages={getattr(skel, bone).languages}) => {skel[bone]!r}",
                )

        return changes if exists else -1



"""
Tools for copying data from ViUR 1.x to ViUR 3.x systems by using the JSON interfaces.
"""

from .importable import Importable
from .importer import Importer

__all__ = [
    "Importable",
    "Importer",
]



import typing as t  # noqa
from datetime import datetime as dt, timedelta as td, timezone as tz  # noqa

from viur.core import utils

__all__ = ["CachedProperty"]

Seconds: t.TypeAlias = int | float
Args = t.ParamSpec("Args")
Value = t.TypeVar("Value")


class CachedProperty(t.Generic[Value, Args]):
    """Wrapper to Cache the result of a function-call"""

    __slots__ = ("lifetime", "func", "args", "_value", "_lifetime_ends")

    def __init__(
        self,
        lifetime: td | Seconds,
        func: t.Callable[Args, Value],
        args: t.Optional[Args.args] = None,
    ):
        """Initiate a new CachedProperty

        :param lifetime: Specifies in seconds how long the cache value should be valid
        :param func: The function that calculates the value
        :param args: Optional Arguments for the function
        """
        if not callable(func):
            raise TypeError("Argument *func* must be a callable function!")
        if args is not None and not isinstance(args, (tuple, list)):
            raise TypeError("Argument *args* must be a tuple, list or None!")
        super(CachedProperty, self).__init__()
        self.lifetime: td = utils.parse.timedelta(lifetime)
        self.func = func
        self.args: Args.args = tuple() if args is None else args
        self._value: Value | None = None
        self._lifetime_ends: dt | None = None

    def get(self) -> Value:
        """Return the value of Property.
        Might be cached or freshly re-calculated."""
        if self._value is not None and utils.utcNow() < self._lifetime_ends:
            return self._value
        self._value = self.func(*self.args)
        self._lifetime_ends = utils.utcNow() + self.lifetime
        return self._value



"""Callable tasks (tasks which can be called by a user from vi-admin)"""
import logging
import typing as t

from viur.core import conf, current, email, translate
from viur.core.bones import FileBone, SelectBone
from viur.core.bones.file import ensureDerived
from viur.core.skeleton import BaseSkeleton, SkeletonInstance, listKnownSkeletons, skeletonByKind
from viur.core.tasks import CallableTask, CallableTaskBase, QueryIter
from .checks import user_has_access

__all__ = [
    "BuildDerivationsDispatcher",
    "BuildDerivations",
]

logger = logging.getLogger(__name__)

_CustomDataType = t.TypedDict("_CustomDataType", {
    "notify": str | None,
    "module": str,
})


@CallableTask
class BuildDerivationsDispatcher(CallableTaskBase):
    """
    This tasks dispatches BuildDerivations with a given module.
    """
    key = "build_derivations"
    name = "Build Derivation"
    descr = "This task can be called to build missing derivations."

    def canCall(self) -> bool:
        """Checks wherever the current user can execute this task"""
        return user_has_access("root")

    def dataSkel(self) -> BaseSkeleton:
        modules = ["*"] + listKnownSkeletons()
        modules.sort()
        skel = BaseSkeleton().clone()
        skel.module = SelectBone(descr="Module", values={x: translate(x) for x in modules}, required=True)
        return skel

    def execute(self, module: str) -> None:
        usr = current.user.get()
        if not usr:
            logger.warning("Don't know who to inform after rebuilding finished")
            notify = None
        else:
            notify = usr["name"]

        if module == "*":
            modules = listKnownSkeletons()
        else:
            modules = [module]
        for module in modules:
            logger.info(f"Rebuilding search index for {module=}")
            self._run(module, notify)

    @staticmethod
    def _run(module: str, notify: str | None) -> None:
        skel_cls = skeletonByKind(module)
        if not skel_cls:
            logger.error(f"{BuildDerivationsDispatcher.__name__}: Invalid {module=}")
            return
        BuildDerivations.startIterOnQuery(skel_cls().all(), {"notify": notify, "module": module})


class BuildDerivations(QueryIter):
    """Task which calls ensureDerived on every FileBone in a Skeleton"""

    @classmethod
    def handleEntry(cls, skel: SkeletonInstance, customData: _CustomDataType) -> None:
        for bone_name, bone_instance in skel.items():
            if not isinstance(bone_instance, FileBone):
                continue
            cls._bone_ensure_derived(skel, bone_instance, bone_name)

    @classmethod
    def _bone_ensure_derived(cls, skel: SkeletonInstance, bone_instance: FileBone, bone_name: str) -> None:
        """Logic from FileBone.postSavedHandler"""

        def handleDerives(values: dict | list) -> None:
            if isinstance(values, dict):
                values = [values]
            for val in values:  # Ensure derives getting build for each file referenced in this relation
                ensureDerived(val["dest"]["key"], f"{skel.kindName}_{bone_name}", bone_instance.derive)

        values = skel[bone_name]
        if bone_instance.derive and values:
            if isinstance(values, dict) and "dest" not in values:  # multi lang
                for lang in values:
                    handleDerives(values[lang])
            else:
                handleDerives(values)

    @classmethod
    def handleFinish(cls, total_count: int, custom_data: _CustomDataType) -> None:
        super().handleFinish(total_count, custom_data)
        if not custom_data["notify"]:
            return
        txt = (
            f"{conf.instance.project_id}: Build derivations finished for {custom_data['module']}\n\n"
            f"ViUR finished to build derivations for module {custom_data['module']}.\n"
            f"{total_count} records updated in total on this kind."
        )
        try:
            email.sendEMail(dests=custom_data["notify"], stringTemplate=txt, skel=None)
        except Exception as exc:  # noqa; OverQuota, whatever
            logger.exception(f'Failed to notify {custom_data["notify"]}')
